#include <p18f1320.h>//strangely enough, it will compile without including this, but it doesn't recognize the LAT bits.
#include <timers.h>
#include <pwm.h>
#include <usart.h>
#include <portb.h>
#include <delays.h>
#include <stdlib.h>
#include <math.h>
#pragma config WDT = OFF, OSC = INTIO1, FSCM = OFF, IESO = OFF, LVP = OFF, MCLRE = OFF, STVR = OFF, DEBUG = OFF, WRT0 = OFF, WRT1 = OFF, CP0 = OFF, CP1 = OFF, CPD = OFF, BOR = OFF

#define SPACE_BETWEEN_MICS 900
#define CLAPS_TO_USE 5



//pins connected to the first step motor's controller. It's important to use the LAT bits instead of PORT for outputs because using PORT can cause problems if you write to one pin right after another.
#define MOTOR1_BIAS_SET LATAbits.LATA3
#define MOTOR1_CLOCK LATAbits.LATA0
#define MOTOR1_NOTCW_CCW LATAbits.LATA1


//pins connected to the second step motor's controller
//I was going to use RA5, but apparently it can only be an input. Something to remember.
#define MOTOR2_BIAS_SET LATBbits.LATB7
#define MOTOR2_CLOCK LATBbits.LATB5
#define MOTOR2_NOTCW_CCW LATAbits.LATA2


static char outputstring[6];
static int times[CLAPS_TO_USE];
static int angle;
static int sum;





void listen(void)
{	//variables have to all be declared at the beginning of a function in C, which is very annoying.
	static char clap;//which clap we're on. Used as index into times[] array
	static char negativeifleftside;//will be -1 if left mic (RB0) heard sound first, 1 if the right did.
	clap = 0;
	while(clap < CLAPS_TO_USE)
	{	
		INTCON3bits.INT2IF = 0;//clear the interrupt flag so it can be set.
		INTCONbits.INT0IF = 0;//clear the interrupt flag so it can be set.
		WriteTimer0(0);//this is used for timing out
		INTCONbits.TMR0IF = 0;//I needed a way to make this loop reset to clap 0 after a certain amount of time.
							//This bit gets set to 1 when the timer overflows, which will happen after 2 seconds.
							//I'm using this the same way as the other interrupt flags-- interrupts are off, but I poll the IF bit.
							//This saves me from having to read the word in timer0. Word ops take a lot of instructions, which would
							//interfere with accurate timing of the claps. This way, when this bit goes high, I know to reset everything.
		
		while( !(INTCONbits.INT0IF || INTCON3bits.INT2IF || INTCONbits.TMR0IF) )//it will stay in this loop until one of the interrupt flags goes high, or it times out (from TMR0IF).
			{}
		WriteTimer1(0);		//set timer to 0.
		
		
		if(INTCONbits.INT0IF == 1)//the left mic's comparator output should go in INT0
			negativeifleftside = -1;
		else
			negativeifleftside = 1;
		
		while( !((INTCONbits.INT0IF && INTCON3bits.INT2IF) || INTCONbits.TMR0IF) )//it will exit loop when both of the interrupt flags are high, or it times out
			{}
		
		//at this point, the second flag has just gone high, meaning the second mic just heard the sound.
		times[clap] = ReadTimer1();
		times[clap] = times[clap] * negativeifleftside;//A negative time will mean it's the left side.
														//this saves me from having to make 2 arrays.
		putrsUSART("clap: ");
		putsUSART(itoa(clap, outputstring) );
		putrsUSART(" has time: ");
		putsUSART(itoa(times[clap], outputstring) );
		clap++;
		if((clap < CLAPS_TO_USE) && !(INTCONbits.TMR0IF))//if this isn't the last one, and we aren't here because of a timeout
			Delay10KTCYx(40);//delay 400 thousand cycles = 200 thousand uS = .2 second for echoes to die out.
		
		if(INTCONbits.TMR0IF)
			clap = 0;//if the timer overflowed, then no one clapped for 2 seconds, so the next clap should be regarded as part of a new series of claps.
	}
	return;
}










//returns the absolute value. That function didn't seem to be in the math library; I may have missed it.
int abs(int x)
{
	if(x<0)
		return -1*x;
	else
		return x;
}




//calculate
//uses the global times[] array. Finds the times that agree with each other, averages them, finds the angle.
void calculate(void)
{
	static char i,j;
	static char valid;
	static char totalvalid;
	static char validsdoexist;//gets set to 1 if we find any valid times.
	validsdoexist = 0;
	totalvalid = 0;
	sum=0;	
	//this part finds the values that agree with each other and takes their average.
	for(i=0;i<CLAPS_TO_USE;i++)
	{
		valid=0;//for each location, start out assuming it's inaccurate
		for(j=0;j<CLAPS_TO_USE;j++)
		{
			if( (times[i]+80 > times[j]) && (times[i]-80 < times[j]) && (i!=j) )
			{
				valid = 1;
				validsdoexist = 1;
			}
			//else, valid stays as it was. The idea of this is that, based on having the PIC tell me times
			//using the PIC serial port, my PC's serial port, and the COMtester program, I saw that invalid
			//times tended to be random big numbers, and valid times were small numbers (like 0 to 1000)
			//which were pretty close to each other, like within +- 80. So what these loops do is, for each
			//recorded time, they compare it to all others. If it agrees with any of the others, it's deemed
			//valid. I realize that if 2 invalid numbers happened to agree, that would be a problem, but that's unlikely,
			//and if it did happen, there would be no easy way to deal with it anyway, so I'll accept the fact
			//that the device might occasionally fling candy off in a random direction.
		}//when this loop is over, valid will be 1 if any of the other times match this one.
		if (valid == 1)
		{
			sum = sum + times[i];//there's no chance of overflow with typical times
			totalvalid++;
		}
	}
	sum = sum / totalvalid;//"sum" variable now actually holds average.
	
	//if no times agree, or if they're too big, this will go with the smallest one
	if( (validsdoexist == 0) || (sum > SPACE_BETWEEN_MICS) )
	{//the error values usually are larger than the true one, so pick the min
		sum = times[0];
		for(i=1;i<CLAPS_TO_USE;i++)
		{
			if(abs(times[i]) < abs(sum))//the most reliable numbers are the ones closest to 0.
				sum = times[i];
			//else, keep sum at the most recent value because that's still the min. found.
		}//after the loop, sum will equal the min.
	}	

	
	
	
	
	////////////////////////now, we have the time to use. Calculate the angle to target from it.
	//if the value of sum is still too big, then there were no valid times.
	if(abs(sum) > SPACE_BETWEEN_MICS)
	{
		sum = 0;
	}
	putrsUSART("chosen: ");
	putsUSART(itoa(sum, outputstring) );
	//now calculate the angle. Theta = ArcCos ( a / h ). Measured time = a, known time separation of mics = h.

	angle = (int) ( asin(((float)sum)/900) * 57.3 );

	putrsUSART("angle: ");
	putsUSART(itoa(angle, outputstring) );
	//now the angle is in "angle" and the direction is in its sign
}







void fire( void )
{
	if(angle < 0)//turn left
	{
		MOTOR1_NOTCW_CCW = 0;
		angle = angle * -1;//make positive so I can use it as a step counter
	}
	else//turn right
		MOTOR1_NOTCW_CCW = 1;

	MOTOR1_BIAS_SET = 1;//will turn on the motor controller's outputs.
	//325 steps will rotate my turret once. So to convert "angle" to steps, do angle * 325 steps/ 360 degrees.
	//the max. value of angle is 90, so there's no chance of overflow.
	//angle = (angle * 325) / 360;//increased some to compensate for a mechanical issue
	//now angle holds the number of steps to take with the step motor.
	angle = angle * 2;//in the loop below, angle is decremented twice for each rising edge.
	while(angle > 0 )
	{
		Delay10KTCYx(3);//the controller steps on the rising edge, so step time
						//is twice that. This was the quickest step speed that still made the motor move
						//the turret.
		MOTOR1_CLOCK = !MOTOR1_CLOCK;//this connects to controller's clock input.
		angle--;
	}
	//now turret is aimed, so fire
	
	//move bolt forward
	angle = 80;//just used as a counter.
	MOTOR2_BIAS_SET = 1;//tells controller to send current through motor. This pin controls an NPN transistor which controls the BIAS/SET input
	MOTOR2_NOTCW_CCW = 0;
	while(angle > 0)
	{
		Delay10KTCYx(4);
		MOTOR2_CLOCK = ! MOTOR2_CLOCK;
		angle--;
	}
	
	MOTOR1_BIAS_SET = 0;//no need to hold steady any more, so cut power to aiming motor.
	//now move it back
	angle = 80;//just used as a counter.
	MOTOR2_NOTCW_CCW = 1;
	while(angle > 0)//move the bolt back by the same amount
	{
		Delay10KTCYx(4);
		MOTOR2_CLOCK = ! MOTOR2_CLOCK;
		angle--;
	}
	MOTOR2_BIAS_SET = 0;//stop powering motor to save power
	Delay10KTCYx(200);//delay 1 second so mics don't hear noise from spinning motors
}











//high_isr ---not used in this program, left here in case I decide to change that.

#pragma interrupt high_isr
void high_isr(void)
{
//left here in case I need it
}



 //* For PIC18 devices the high interrupt vector is found at
 //* 00000008h. The following code will branch to the
 //* high interrupt service routine function to handle
 //* interrupts that occur at the high vector.
 //In the source code, this must be located after the high_isr function.
#pragma code high_vector=0x08
void interrupt_at_high_vector(void)
{
  _asm GOTO high_isr _endasm
}
#pragma code // return to the default code section
















void main (void)
{	//general initialization code
	OSCCON = 0x70;	//set it to use the 8MHz internal clock.
	ADCON1 = 0x7F;	//set all ADC-capable pins to NOT be analog inputs, just digital IO.
	//end of initialization code

	//application-specific init code
	DDRB=0x0C;  //all outputs except rb0 & rb2. Those are the inputs from microphone comparators.
  	DDRBbits.RB2=1;
  	DDRBbits.RB0=1;
  	DDRA = 0;
  	PORTA=0;
  	PORTB=0;
	MOTOR2_BIAS_SET = 0;	//1 to send current, 0 to turn off.
  	MOTOR1_BIAS_SET = 0;


						
	OpenTimer1( TIMER_INT_OFF    &
            T1_16BIT_RW      &
            T1_SOURCE_INT   &
            T1_PS_1_1       &
            T1_OSC1EN_OFF   &
            T1_SYNC_EXT_OFF );//Timer1 will be used for measuring time delay between mics
            //Timer period will be 1 * (instruction cycle time of .5uS)
						//= .5 uS with the 8Mhz oscillator.	
	
	OpenTimer0(TIMER_INT_OFF & 
				T0_16BIT &
				T0_SOURCE_INT &
				T0_PS_1_64);//Timer0 will be used to "time out" and reset to 0 claps after ~2 seconds of silence.
							//With these settings, the timer will overflow after ~2.1 seconds. That sets a flag, which
							//provides an easy way to check if 2 seconds have passed. (And it takes less code.)
	
	
	//interrupt config. The PIC's interrupt flags are used in the listen() function.
	OpenRB2INT(PORTB_CHANGE_INT_ON &
				RISING_EDGE_INT &
				PORTB_PULLUPS_ON);

	OpenRB0INT(PORTB_CHANGE_INT_ON &
				RISING_EDGE_INT &
				PORTB_PULLUPS_ON);
				
	INTCONbits.GIE=0;	//global interrupt control. Leave interrupts disabled, so no function is actually called.
						//I'm really just using the polling method, but the interrupt flag bits will latch high, so I won't
						//miss a very brief high voltage. If I did normal polling, I'm afraid a high voltage might come and
						//go in the time between port reads.
  	//end of init code
  	
	OpenUSART(USART_TX_INT_OFF &
  			USART_RX_INT_OFF &
  			USART_ASYNCH_MODE &
  			USART_EIGHT_BIT &
  			USART_CONT_RX &
  			USART_BRGH_LOW,
  			12);		//this is the baud rate number. Baudrate with BRGH_LOW is 
  					//Fosc/(64*(baud rate number + 1))
  					//The number 12 should give 9600 baud if clock speed is 8MHz.
  	Delay10KTCYx(200);//will delay 2 million instruction cycles, or 1 second for 8MHz clock
					//(since 4 clock cycles makes 1 instruction cycle).
	putrsUSART("Started up");
	/*
	//the following code makes the Starburst-pushing "bolt" move all the way back, in case it wasn't already.
	angle = 80;//just used as a counter.
	MOTOR2_BIAS_SET = 1;//tells controller to send current through motor
	MOTOR2_NOTCW_CCW = 1;
	while(angle > 0)
	{
		Delay10KTCYx(4);
		MOTOR2_CLOCK = ! MOTOR2_CLOCK;
		angle--;
	}
	MOTOR2_BIAS_SET = 0;//stop powering motor to save power.
	//now the bolt is definitely all the way back, ready to push the next Starburst forward.
	*/ //I decided to just make sure it's in the right position before turning it on.
	
		
	while(1)
	{	
		listen();//waits for claps, records time between them.
		calculate();//looks at times array (a global variable) and determines angle.
		fire();//aims the turret with one step motor, then turns on the firing motors, then feeds
				//a Starburst in to them with the other step motor.
	}
}
